//============================================================================
//
//  Parallel checker
//
//  NoteCreaor.java
//
//  Copyright (c) 2019 Fumiaki Hosaka.
//
//  Permission to copy, use, modify, sell and distribute this software
//  is granted provided this copyright notice appears in all copies.
//  This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//============================================================================
package parallel;

import java.io.*;
import java.util.*;

import static parallel.Constant.*;

public class NoteCreator
{
    public ArrayList<Note[]> tracks; // XML解析後のTrackの音符データ.
    public int useTracks;  // 使用Track数.
    public int lastPos;    // 最終位置.(Tick単位)
    public int beats;      // 拍数.
    public int beatType;   // 拍の単位.

    // 2Track間での発音タイミングのデータ(トラック組み合わせ).
    public ArrayList<ArrayList<OnPointPair>> result;

    public NoteCreator()
    {
    }

    // ・最小Tickの連続として音を配列につめ直す.
    // ・休符を前の音で埋める.
    // ・全Trackの組み合わせで、
    //   2Track間で発音が変化した場所のペアを作成する.
    public void create(ArrayList<Object[]> parts)
    {
        createTrack(parts);
        paddingRest();
        createCombination();
    }

    public void prepare()
    {
        result = new ArrayList<ArrayList<OnPointPair>>();

        // 基本拍を24分割したものを1Tickとし、
        // 1Trackにつき400Tick(4/4で100小節分)を準備する.
        // Trackは分割用に4x2=8としておく.
        tracks = new ArrayList<Note[]>();
        for(int i = 0;i < TRACK_MAX; ++i)
        {
            Note [] t = new Note[BEAT_MAX * CELL_SIZE];
            tracks.add(t);
            for(int j = 0;j < BEAT_MAX * CELL_SIZE; ++j)
                t[j] = null;
        }
    }

    HashMap<String, Integer> trackIdxMap = new HashMap<String, Integer>();

    private void calcTrackCount(ArrayList<Object[]> parts)
    {
        int ttl = 0;
        for(int i=0; i< parts.size(); ++i)
        {
            TreeSet<String> set = new TreeSet<String>();
            Object[] pp = parts.get(i);
            ArrayList<Measure> ml  = (ArrayList<Measure>)pp[1];
            for(int j = 0; j < ml.size(); ++j)
            {
                Measure bar = ml.get(j);
                for(String voice : bar.voiceSet)
                    set.add(voice);
            }

            if (set.size() == 0)
            {
                trackIdxMap.put(String.format("%d:0", i), Integer.valueOf(ttl));
                if (traceOn)
                    System.err.printf("track=%d voice=Non idx=%d\r\n",
                                         i, ttl);
                ttl += 1;
            }
            else
            {
                int k = ttl;
                for(String s : set)
                {
                    trackIdxMap.put(String.format("%d:%s", i, s), Integer.valueOf(k));
                    if (traceOn)
                        System.err.printf("track=%d voice[%s] idx=%d\r\n",
                                            i, s, k);

                    ++k;
                }
                ttl += set.size();
            }
        }
    }

    private void createTrack(ArrayList<Object[]> parts)
    {
        calcTrackCount(parts);

        useTracks = trackIdxMap.size();
        beatType = -1;
        beats    = -1;

        int last  = -1;
        int bType = -1;

        for(int i=0; i< parts.size(); ++i)
        {
            Object[] pp = parts.get(i);
            if (traceOn)
                System.err.printf("[track %d] %s\r\n", i+1, pp[0].toString());

            int pos = 0;
            int barPos = 0;
            int divisions = 0;

            ArrayList<Measure> ml  = (ArrayList<Measure>)pp[1];
            for(int j = 0; j < ml.size(); ++j)
            {
                Measure bar = ml.get(j);
                bar.totalTick = pos;
                if (traceOn)
                {
                    System.err.printf("---%02d---(%d/%d) division = %d  totalTick = %d\r\n",
                                                     bar.no, bar.beats, bar.beatType,
                                                     bar.divisions, bar.totalTick);
                }

                if (bar.beats != 0 && beatType != bar.beatType)
                {
                    if (i == 0)
                    {
                        beats    = bar.beats;
                        beatType = bar.beatType;
                    }
                    bType    = bar.beatType;

                }
                if (bar.divisions != 0 && divisions != bar.divisions)
                {
                    divisions = bar.divisions;
                }

                for(int k = 0; k < bar.notes.size(); ++k)
                {
                    XMLNote nn = bar.notes.get(k);
                    int alt = Integer.parseInt(nn.alter);
                    int oct = Integer.parseInt(nn.octave);
                    int dur = Integer.parseInt(nn.dur);

                    int unit = 24 / divisions;
                    dur = dur * unit;

                    if (nn.isBackup)
                    {
                        pos -= dur;
                        System.err.printf("backup dur=(%s)%d\r\n",
                                           nn.dur, dur);
                        continue;
                    }

                    if (traceOn)
                    {
                        System.err.printf("voice=%s pitch=(%s, %s, %s) chord=%s dur=(%s)%d\r\n",
                                           nn.voice, nn.step, nn.octave, nn.alter,
                                           nn.isChord, nn.dur, dur);
                    }

                    int idx = trackIdxMap.get(String.format("%d:%s", i, nn.voice));

                    if (nn.isChord)
                        setNote(bar, idx, nn.step, alt, oct, pos, dur);
                    else
                        pos = setNote(bar, idx, nn.step, alt, oct, pos, dur);
                }
                if (pos > last)
                    last = pos;
            }
        }
        lastPos = last;
    }

    // step  = "C"など.
    // alter = -1(フラット), 0(ナチュラル), 1(シャープ).
    // octave = 0-8.
    private int setNote(Measure bar, int trackNo, String step, int alter, int octave,
                        int startPos, int duration)
    {
        Note note;
        if (step.equals("-1"))
            note = new Note(trackNo, bar); // rest!
        else
            note = new Note(trackNo, bar, step, alter, octave);

        // Trackの指定位置に長さ分Noteを埋める(Noteは共有).
        Note[] notes = tracks.get(trackNo);
        for(int i = 0; i < duration; ++i)
            notes[ startPos + i ] = note;
        return startPos + duration;
    }

    // 休符を前の音で埋める.
    private void paddingRest()
    {
        if (traceOn)
            System.err.printf(">>>Padding Rest\r\n");

        int chkDur = Options.getCheckDuration(this);

        for(int i = 0;i < useTracks; ++i)
        {
            Note [] t1 = tracks.get(i);

            Note before = null;
            int  restCnt = 0;
            for(int j = 0; j < lastPos; ++j)
            {
                Note curr = t1[j];
                if (curr == null)
                    continue;
                if (curr.no < 0 )
                {
                    if (before == null)
                        continue;
                    restCnt++;
                    if (restCnt > chkDur)
                        continue;
                    // 共有を切るため、新しいInstanceを設定.
                    Note pad = new Note(curr.track, curr.bar);
                    t1[j] = pad;
                    pad.no  = before.no;
                    pad.sig = before.sig;
                    pad.padding = true;
                }
                else
                {
                    before = curr;
                    restCnt = 0;
                }
            }
        }
    }

    // Trackの組み合わせを作成.
    private void createCombination()
    {
        for(int i = 0;i < useTracks; ++i)
        {
            for(int j = i + 1;j < useTracks; ++j)
            {
                result.add(createOnPoints(i, j));
            }
        }
    }

    private ArrayList<OnPointPair> createOnPoints(int tno1, int tno2)
    {
        ArrayList<OnPointPair> onpts = new ArrayList<OnPointPair>();
        Note [] t1 = tracks.get(tno1);
        Note [] t2 = tracks.get(tno2);

        int bef1 = -100;
        int bef2 = -100;

        for(int i = 0; i < lastPos; ++i)
        {
            Note w1 = t1[i];
            Note w2 = t2[i];

            // どちらかが休符の場合は継続.
            if (w1 == null ||  w2 == null)
                continue;
            if (w1.no == -1 ||  w2.no == -1)
                continue;

            int df = (w1.no > w2.no) ? w1.no - w2.no
                                     : w2.no - w1.no;

            if (df != 12)
                df = df % 12;

            // 音が変更されていたら結果に追加.
            if (bef1 != w1.no || bef2 != w2.no)
            {
                onpts.add(new OnPointPair(i, df, w1, w2));
                bef1 = w1.no;
                bef2 = w2.no;
            }
        }
        dumpOnPoint(onpts);
        return onpts;
    }

    void dumpOnPoint(ArrayList<OnPointPair> onpts)
    {
        if (!traceOn)
            return;

        System.err.printf("==================\r\n");
        for(int i=0; i < onpts.size(); ++i)
        {
            OnPointPair op = onpts.get(i);
            System.err.printf("[%d] %s\r\n", i, op.toString());
        }
    }

}
